# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import Callable, Dict, List, Optional, Tuple

from tqdm import tqdm
from wakepy import keep

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.geometries import Geometry
from ansys.simai.core.data.types import (
    Identifiable,
    NamedFile,
    get_id_from_identifiable,
)
from ansys.simai.core.data.workspaces import Workspace
from ansys.simai.core.errors import InvalidArguments

logger = logging.getLogger(__name__)


class Optimization(ComputableDataModel):
    """Provides the local representation of an optimization definition object."""

    def _try_geometry(
        self, geometry: Identifiable[Geometry], geometry_parameters: Dict
    ) -> "OptimizationTrialRun":
        return self._client._optimization_trial_run_directory._try_geometry(
            self.id, geometry, geometry_parameters
        )


class OptimizationTrialRun(ComputableDataModel):
    """Provides the local representation of an optimization trial run object.

    The optimization trial run is an iteration of the optimization process.
    Each trial run tests a geometry and returns new parameters for the next geometry to try.
    """


class OptimizationDirectory(Directory[Optimization]):
    """Provides a collection of methods related to optimizations.

    This class is accessed through ``client.optimizations``.

    Example:
        .. code-block:: python

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.optimizations.run(...)
    """

    _data_model = Optimization

    def get(self, optimization_id: str) -> Optimization:
        """Get a specific optimization object from the server.

        Args:
            optimization_id: ID of the optimization.

        Returns:
            :py:class:`Optimization`.
        """
        return self._model_from(self._client._api.get_optimization(optimization_id))

    def run(
        self,
        geometry_generation_fn: Callable[..., NamedFile],
        geometry_parameters: Dict[str, Tuple[float, float]],
        boundary_conditions: Dict[str, float],
        n_iters: int,
        minimize: Optional[List[str]] = None,
        maximize: Optional[List[str]] = None,
        outcome_constraints: Optional[List[str]] = None,
        show_progress: bool = False,
        workspace: Optional[Identifiable[Workspace]] = None,
    ) -> List[Dict]:
        """Run an optimization process.

        Args:
            geometry_generation_fn: Function to call to generate a new geometry
                with the generated parameters. This parameter should return a
                :obj:`~ansys.simai.core.data.types.NamedFile` object.
            geometry_parameters: Name of the geometry parameters and their bounds or possible values (choices).
            boundary_conditions: Values of the boundary conditions to perform the optimization at.
                The values should map to existing boundary conditions in your project/workspace.
            minimize: List of global coefficients to minimize.
                The global coefficients should map to existing coefficients in your project/workspace.
            maximize: List of global coefficients to maximize.
                The global coefficients should map to existing coefficients in your project/workspace.
            outcome_constraints: List of strings representing a linear inequality constraint
                on a global coefficient. The outcome constraint should be in the form ``gc >= x``,
                where:

                - ``gc`` is a valid global coefficient name.
                - ``x`` is a float bound.
                - The comparison operator is ``>=`` or ``<=``.

            n_iters: Number of iterations of the optimization loop.
            show_progress: Whether to print progress on stdout.
            workspace: Workspace to run the optimization in. If a workspace is
                not specified, the default is the configured workspace.

        Returns:
            List of dictionaries representing the result of each iteration. when constraints
            are specified, the list can be shorter than the number of iterations.

        Warning:
            This is a long-running process and your computer must be powered on to generate the iterations.
            This method attempts to prevent your computer from sleeping, keeping your computer open
            during the process.

        Example:
          .. code-block:: python

            import ansys.simai.core


            # Function takes the parameters
            def my_geometry_generation_function(param_a, param_b):
                # Implementation
                return "/path/to/generated/geometry.stl"


            simai = ansys.simai.core.from_config(workspace="optimization-workspace")

            results = simai.optimizations.run(
                geometry_generation_fn=my_geometry_generation_function,
                geometry_parameters={
                    "param_a": {"bounds": (-12.5, 12.5)},
                    "param_b": {"choices": (0, 1)},
                },
                minimize=["TotalForceX"],
                boundary_conditions={"VelocityX": 10.5},
                outcome_constraints=["TotalForceX <= 10"],
                n_iters=100,
            )

            print(results)
        """
        workspace_id = get_id_from_identifiable(workspace, True, self._client._current_workspace)
        if not minimize and not maximize:
            raise InvalidArguments("No global coefficient to optimize.")
        objective = {}
        if minimize:
            for global_coefficient in minimize:
                objective[global_coefficient] = {"minimize": True}
        if maximize:
            for global_coefficient in maximize:
                objective[global_coefficient] = {"minimize": False}
        optimization_parameters = {
            "boundary_conditions": boundary_conditions,
            "geometry_parameters": geometry_parameters,
            "n_iters": n_iters,
            "objective": objective,
            "outcome_constraints": outcome_constraints or [],
        }
        with tqdm(total=n_iters, disable=not show_progress) as progress_bar:
            progress_bar.set_description("Creating optimization definition.")
            optimization = self._model_from(
                self._client._api.define_optimization(workspace_id, optimization_parameters)
            )
            optimization.wait()
            geometry_parameters = optimization.fields["initial_geometry_parameters"]
            logger.debug("Optimization defined. Starting optimization loop.")
            iterations_results: List[Dict] = []
            with keep.running(on_fail="warn") as k:
                if not k.success:
                    logger.info("Failed to get sleep inhibition lock.")
                while geometry_parameters:
                    logger.debug(f"Generating geometry with parameters {geometry_parameters}.")
                    progress_bar.set_description("Generating geometry.")
                    # TODO: Somehow keep session alive for long geometry generation
                    generated_geometry = geometry_generation_fn(**geometry_parameters)
                    logger.debug("Uploading geometry.")
                    progress_bar.set_description("Uploading geometry.")
                    # TODO: Name geometry ourselves ? Then we need to know the output format
                    geometry = self._client.geometries.upload(
                        generated_geometry,
                        metadata=geometry_parameters,
                        workspace_id=workspace_id,
                    )
                    logger.debug("Running trial.")
                    progress_bar.set_description("Running trial.")
                    trial_run = optimization._try_geometry(geometry, geometry_parameters)
                    trial_run.wait()
                    iteration_result = {
                        "parameters": geometry_parameters,
                        "objective": trial_run.fields["outcome_values"],
                    }
                    progress_bar.set_postfix(**iteration_result)
                    if trial_run.fields.get("is_feasible", True):
                        iterations_results.append(iteration_result)
                    else:
                        logger.debug("Trial run results did not match constraints. Skipping.")
                    geometry_parameters = trial_run.fields["next_geometry_parameters"]
                    logger.debug("Trial complete.")
                    progress_bar.update(1)
                logger.debug("Optimization complete.")
                progress_bar.set_description("Optimization complete.")
            return iterations_results


# Undocumented for now, users don't really need to interact with it
class OptimizationTrialRunDirectory(Directory[OptimizationTrialRun]):
    _data_model = OptimizationTrialRun

    def get(self, trial_run_id: str):
        """Get a specific trial run from the server."""
        return self._model_from(self._client._api.get_optimization_trial_run(trial_run_id))

    def _try_geometry(
        self,
        optimization: Identifiable[Optimization],
        geometry: Identifiable[Geometry],
        geometry_parameters: Dict,
    ) -> OptimizationTrialRun:
        geometry_id = get_id_from_identifiable(geometry)
        optimization_id = get_id_from_identifiable(optimization)
        return self._model_from(
            self._client._api.run_optimization_trial(
                optimization_id, geometry_id, geometry_parameters
            )
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import threading
import time
import webbrowser
from datetime import datetime, timedelta
from typing import Optional
from urllib.parse import urljoin

import requests
from pydantic import BaseModel, ValidationError
from requests.auth import AuthBase

from ansys.simai.core.errors import ConnectionError
from ansys.simai.core.utils.configuration import ClientConfig, Credentials
from ansys.simai.core.utils.files import get_cache_dir
from ansys.simai.core.utils.requests import handle_response

logger = logging.getLogger(__name__)


class _AuthTokens(BaseModel):
    """Represents the OIDC tokens received from the auth server.

    The class can fetch and refresh these tokens.
    It caches the tokens to disk automatically.
    """

    _EXPIRATION_BUFFER = timedelta(seconds=5)

    cache_file: Optional[str] = None
    access_token: str
    expires_in: int
    refresh_expires_in: int
    expiration: datetime = None
    refresh_expiration: datetime = None
    refresh_token: str
    # ...  (unused fields were removed for simplicity)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.expiration is None:
            now = datetime.utcnow()
            self.expiration = now + timedelta(seconds=self.expires_in)
            self.refresh_expiration = now + timedelta(seconds=self.refresh_expires_in)
        if self.cache_file is not None:
            with open(self.cache_file, "w") as f:
                f.write(self.model_dump_json())

    @classmethod
    def from_cache(cls, file_path: Optional[str]) -> "Optional[_AuthTokens]":
        if file_path is None:
            return None
        try:
            auth = cls.parse_file(file_path)
            # don't return expired auth tokens
            if not auth.is_refresh_token_expired():
                return auth
        except (IOError, json.JSONDecodeError) as e:
            logger.debug(f"Could not read auth cache: {e}")
        except ValidationError as e:
            logger.warning(f"ValidationError while reading auth cache: {e}")
        return None

    @classmethod
    def from_request_direct_grant(
        cls, session: requests.Session, token_url, creds: Credentials
    ) -> "_AuthTokens":
        logger.debug("Getting authentication tokens.")
        request_params = {
            "client_id": "sdk",
            "grant_type": "password",
            "scope": "openid",
            **creds.model_dump(),
        }
        return cls(**handle_response(session.post(token_url, data=request_params)))

    @classmethod
    def from_request_device_auth(
        cls, session: requests.Session, device_auth_url, token_url, cache_file=None
    ) -> "_AuthTokens":
        auth_codes = handle_response(
            session.post(device_auth_url, data={"client_id": "sdk", "scope": "openid"})
        )
        print(  # noqa: T201
            f"Go to {auth_codes['verification_uri']} and enter the code {auth_codes['user_code']}"
        )
        webbrowser.open(auth_codes["verification_uri_complete"])
        while True:
            validation = session.post(
                token_url,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "client_id": "sdk",
                    "device_code": auth_codes["device_code"],
                },
            )
            if b"authorization_pending" not in validation.content:
                return cls(**handle_response(validation), cache_file=cache_file)
            # polling can't be faster or auth server returns HTTP 400 Slow down
            time.sleep(5)

    def is_token_expired(self):
        return datetime.utcnow() > (self.expiration - self._EXPIRATION_BUFFER)

    def is_refresh_token_expired(self):
        return datetime.utcnow() > (self.refresh_expiration - self._EXPIRATION_BUFFER)

    def refresh(self, session: requests.Session, token_url: str) -> "_AuthTokens":
        logger.debug("Refreshing authentication tokens.")
        request_params = {
            "client_id": "sdk",
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
        }
        try:
            return _AuthTokens(
                **handle_response(session.post(token_url, data=request_params)),
                cache_file=self.cache_file,
            )
        except requests.exceptions.ConnectionError as e:
            raise ConnectionError(e) from None


def _get_cached_or_request_device_auth(
    session: requests.Session,
    device_auth_url: str,
    token_url: str,
    auth_cache_hash: Optional[str],
):
    auth = None
    auth_cache_path = None
    if auth_cache_hash:  # Try to fetch auth from cache
        auth_cache_path = str(get_cache_dir() / f"tokens-{auth_cache_hash}.json")
        auth = _AuthTokens.from_cache(auth_cache_path)
        if auth:
            try:
                auth = auth.refresh(session, token_url)
            except Exception as e:
                logger.debug(f"Auth refresh error: {e}")
                auth = None
    if auth is None:  # Otherwise request the user to authenticate
        auth = _AuthTokens.from_request_device_auth(
            session, device_auth_url, token_url, auth_cache_path
        )
    return auth


class Authenticator(AuthBase):
    def __init__(self, config: ClientConfig, session: requests.Session) -> None:
        self._session = session
        self._enabled = not getattr(config, "_disable_authentication", False)
        if not self._enabled:
            logger.debug("Disabling authentication logic.")
            return
        self._url_prefix = config.url
        # HACK: start with a slash to override the /v2/ on the api url
        self._realm_url = urljoin(str(config.url), "/auth/realms/simai")
        self._token_url = f"{self._realm_url}/protocol/openid-connect/token"
        self._organization_name = config.organization
        self._refresh_timer = None
        if config.credentials:
            auth = _AuthTokens.from_request_direct_grant(
                self._session, self._token_url, config.credentials
            )
        else:
            device_auth_url = f"{self._realm_url}/protocol/openid-connect/auth/device"
            auth_hash = None if config.disable_cache else config._auth_hash()
            auth = _get_cached_or_request_device_auth(
                self._session, device_auth_url, self._token_url, auth_hash
            )
        self._authentication = auth
        self._schedule_auth_refresh()

    def __call__(self, request: requests.Request) -> requests.Request:
        """Call to prepare the requests.

        Args:
            request: Request to authenticate.

        Returns:
            Request with the authentication.
        """
        request_host = request.url.split("://", 1)[-1]  # ignore protocol part
        if self._enabled and request_host.startswith(self._url_prefix.host):
            is_token_expired = self._authentication.is_token_expired()
            # So the token doesn't expire during requests that upload a large amount of data
            is_request_multipart_data = "multipart/form_data" in request.headers.get(
                "Content-Type", ""
            )
            is_auth_request = self._realm_url in request.url
            if not is_auth_request and is_token_expired or is_request_multipart_data:
                self._refresh_auth()
            request.headers["Authorization"] = f"Bearer {self._authentication.access_token}"
            request.headers["X-Org"] = self._organization_name
        return request

    def _refresh_auth(self):
        """Refresh the authentication."""
        self._authentication = self._authentication.refresh(self._session, self._token_url)
        self._schedule_auth_refresh()

    def _schedule_auth_refresh(self):
        """Schedule authentication refresh to avoids refresh token expiring if the client is idle for a long time."""
        if self._refresh_timer:
            self._refresh_timer.cancel()
        self._refresh_timer = threading.Timer(
            self._authentication.refresh_expires_in - 60, self._refresh_auth
        )
        self._refresh_timer.daemon = True
        self._refresh_timer.start()



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from ansys.simai.core.data.types import Identifiable, get_id_from_identifiable

if TYPE_CHECKING:
    from ansys.simai.core.client import SimAIClient
    from ansys.simai.core.data.workspaces import Workspace


class DesignOfExperimentsCollection:
    """Provides a collection of methods related to design of experiments.

    This class is accessed through ``client.design_of_experiments``.
    """

    def __init__(self, client: "SimAIClient"):
        self._client = client

    def download(
        self,
        file: Union[str, Path],
        format: str = "xlsx",
        workspace: Optional[Identifiable["Workspace"]] = None,
    ) -> None:
        """Download the design of experiments data to the specified file or path.

        Args:
            file: Path of the file to put the content into.
            format: Expected format. The default is ``'xlsx'``. Options are ``'xlsx'``
                and ``'csv'``.
            workspace: Workspace to download the design of experiments from. If not
                specified, the current workspace is used.

        Example:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                simai.design_of_experiments.download(file="~/exp_plan.xlsx", format="xlsx")
        """
        workspace_id = get_id_from_identifiable(workspace, default=self._client._current_workspace)
        self._client._api.download_design_of_experiments(file, format, workspace_id)

    def in_memory(
        self, format: Optional[str] = "csv", workspace: Optional[Identifiable["Workspace"]] = None
    ) -> io.BytesIO:
        """Load the design of experiments data in memory.

        Args:
            file: Path of the file to put the content into.
            format: Expected format. The default is ``'csv'``. Options are ``'xlsx'``
                and ``'csv'``.
            workspace: Workspace to download the design of experiments from. If not
                specified, the current workspace is used.

        Returns:
            :class:`~io.BytesIO` object containing the design of experiments data.

        Example:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                workspace = simai.workspaces.list()[0]
                data = simai.design_of_experiments.in_memory(format="csv", workspace=workspace)
                # Read data with CSV reader, ...
        """
        workspace_id = get_id_from_identifiable(workspace, default=self._client._current_workspace)
        return self._client._api.download_design_of_experiments(None, format, workspace_id)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ruff: noqa: F401

from importlib.metadata import version

try:
    __version__ = version("ansys-simai-core")
except Exception:
    __version__ = "n/a"

import ansys.simai.core.errors
from ansys.simai.core.client import SimAIClient, from_config
from ansys.simai.core.data.post_processings import (
    CustomVolumePointCloud,
    GlobalCoefficients,
    Slice,
    SurfaceEvol,
    SurfaceVTP,
    VolumeVTU,
)
